///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2017 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////

// This NABLA port is an implementation of the benchmark LULESH:
// the Livermore Unstructured Lagrangian Explicit Shock Hydrodynamics
// (see https://computation.llnl.gov/casc/ShockHydro/ for details)

with cartesian;


// ****************************************************************************
// * Options
// ****************************************************************************
options{
  ℝ LENGTH                    = 1.125;
  ℕ X_EDGE_ELEMS              = 8;
  ℕ Y_EDGE_ELEMS              = 8;
  ℕ Z_EDGE_ELEMS              = 8;
  
  ℝ option_δt_fixed           = -1.0e-7;  // fixed time increment
  ℝ option_δt_initial         = 1.0e-7;   // variable time increment
  ℝ option_δt_courant         = 1.0e+20;
  ℝ option_δt_hydro           = 1.0e+20;
  ℝ option_δt_mult_lower_b    = 1.1;
  ℝ option_δt_mult_upper_b    = 1.2;
  ℝ option_initial_energy     = 3.948746e+7;
  ℝ option_stoptime           = 1.0e-2;   // end time for simulation
  ℝ option_hgcoef             = 3.0;      // hourglass control
  ℝ option_qstop              = 1.0e+12;  // excessive q indicator
  ℝ option_monoq_max_slope    = 1.0;
  ℝ option_monoq_limiter_mult = 2.0;
  ℝ option_e_cut              = 1.0e-7;   // energy tolerance
  ℝ option_p_cut              = 1.0e-7;   // pressure tolerance
  ℝ option_q_cut              = 1.0e-7;   // q tolerance
  ℝ option_u_cut              = 1.0e-7;   // node velocity cut-off value
  ℝ option_v_cut              = 1.0e-10;  // relative volume tolerance
  ℝ option_qlc_monoq          = 0.5;      // linear term coef for q
  ℝ option_qqc_monoq          = 0.6666666666667;//666666; // quadratic term coef for q
  ℝ option_qqc                = 2.0;
  ℝ option_eosvmax            = 1.0e+9;
  ℝ option_eosvmin            = 1.0e-9;
  ℝ option_pmin               = 0.0;      // pressure floor
  ℝ option_emin               = -1.0e+15; // energy floor
  ℝ option_dvovmax            = 0.1;      // maximum allowable volume change
  ℝ option_refdens            = 1.0;      // reference density
  ℝ option_dtmax              = 1.0e-2;   // maximum allowable time increment
  ℾ option_chaos              = false;
  ℝ option_chaos_seed         = 1.1234567890123;//456789;
  ℕ option_max_iterations     = 32768;
};


// ****************************************************************************
// * Node Variables
// ****************************************************************************
nodes{
  ℝ³ 𝜕x;           // Velocity vector
  ℝ³ 𝜕𝜕x;          // Acceleration vector
  ℝ³ nForce;       // Force vector
  ℝ nodalMass;     // Nodal mass
};


// ****************************************************************************
// * Element Variables
// ****************************************************************************
cells{
  // Table 1.2 Element Variables  ////////////////
  ℝ p;              // pressure
  ℝ e;              // internal energy, (to synchronize)
  ℝ q;              // artificial viscosity
  ℝ v;              // relative volume
  ℝ calc_volume;    // instant relative volume
  ℝ vdov;           // relative volume change per volume
  ℝ delv;           // relative volume change
  ℝ volo;           // reference (initial) volume
  ℝ arealg;         // characteristic length
  ℝ³ ε;             // diagonal terms of deviatoric strain  dxx(),dyy(),dzz()
  ℝ ql;             // artificial viscosity linear term, (to synchronize)
  ℝ qq;             // artificial viscosity quadratic term, (to synchronize)
  ℝ³ cForce[nodes];
  // Temporaries /////////////////////////////////
  ℝ delv_xi;        // velocity gradient
  ℝ delv_eta;
  ℝ delv_zeta;
  ℝ delx_xi;        // coordinate gradient
  ℝ delx_eta;
  ℝ delx_zeta;
  ℝ phixi;
  ℝ phieta;
  ℝ phizeta;
  ℝ vnew;           // new relative volume
  ℝ elemMass;       // mass
  // EoS /////////////////////////////////////////
  ℝ e_old;
  ℝ delvc;
  ℝ p_old;
  ℝ q_old;
  ℝ compression;
  ℝ compHalfStep;
  ℝ work;
  ℝ p_new;
  ℝ e_new;
  ℝ q_new;
  ℝ bvc;
  ℝ pbvc;
  ℝ vnewc;
  ℝ pHalfStep;
  ℝ sound_speed;
  // Boundary Conditions Flags //////////////////////////////////////////////
  ℕ elemBC;          // symmetry/free-surface flags for each elem face
  // Reductions ////////////////////////////
  ℝ δt_cell_hydro;
  ℝ δt_cell_courant;
};


// ****************************************************************************
// * Global Variables
// ****************************************************************************
global{
  ℝ δt_courant;         // Courant time constraint
  ℝ δt_hydro;           // Hydro time constraint
};


// ****************************************************************************
// * Initialization Part @ ]-∞,-0.0[
// ****************************************************************************


// ****************************************************************************
// * ini
// ****************************************************************************
void ini(void) @ -10.0{
  δt=(option_chaos)?option_δt_initial*option_chaos_seed:0.0;
  δt_hydro=option_δt_hydro;
  δt_courant=option_δt_courant;
  //atexit(dumpFinalEnergy);
}

∀ nodes void chaosNodes(void) out (node coord) @ -12.0{
  coord *= (option_chaos)?option_chaos_seed:1.0;
  //const Real chaos = (((Real)uid)+1.0)*option_chaos_seed;
  //coord = (option_chaos)?chaos:1.0;
  //info()<<"coord="<<coord;
  //printf("\ncoord[%d]=(%.21e,%.21e,%.21e)",uid,coord.x,coord.y,coord.z);
}
/*nodes void dumpCorners(void)@-12.0{
  foreach cell{
    const int corner = node_cell_corner[8*n+i];
    if (corner==-1) continue;
    if (uid==31) printf("\n[#%d] node_#%d is corner %d of cell %d",i,uid,node_cell_corner[8*n+i],node_cell[8*n+i]);
  }
  }*/


  
// ****************************************************************************
// * Set up boundary condition information
// * Set up elemement connectivity information
// ****************************************************************************
#define XI_M        0x003
#define XI_M_SYMM   0x001
#define XI_M_FREE   0x002
#define XI_P        0x00C
#define XI_P_SYMM   0x004
#define XI_P_FREE   0x008
#define ETA_M       0x030
#define ETA_M_SYMM  0x010
#define ETA_M_FREE  0x020
#define ETA_P       0x0C0
#define ETA_P_SYMM  0x040
#define ETA_P_FREE  0x080
#define ZETA_M      0x300
#define ZETA_M_SYMM 0x100
#define ZETA_M_FREE 0x200
#define ZETA_P      0xC00
#define ZETA_P_SYMM 0x400
#define ZETA_P_FREE 0x800
∀ cells void iniCellBC(void) in (node coord) out (cell elemBC) @ -9.5{
  const ℝ zero = 0.0;
  const ℝ maxBoundaryX = LENGTH;//X_EDGE_TICK*X_EDGE_ELEMS;
  const ℝ maxBoundaryY = LENGTH;//Y_EDGE_TICK*Y_EDGE_ELEMS;
  const ℝ maxBoundaryZ = LENGTH;//Z_EDGE_TICK*Z_EDGE_ELEMS;
  elemBC=0; // clear BCs by default
  ∀ node{
    elemBC |= (coord.x==zero)?XI_M_SYMM;
    elemBC |= (coord.y==zero)?ETA_M_SYMM;
    elemBC |= (coord.z==zero)?ZETA_M_SYMM;
    elemBC |= (coord.x==maxBoundaryX)?XI_P_FREE;
    elemBC |= (coord.y==maxBoundaryY)?ETA_P_FREE;
    elemBC |= (coord.z==maxBoundaryZ)?ZETA_P_FREE;
  }
  //debug()<<"#"<<uid<<" elemBC="<<elemBC;
  //printf("\n#%d elemBC=%d",uid,elemBC);
}



// ****************************************************************************
// * Cells initialization
// ****************************************************************************
∀ cells void iniCells(void) in (node coord)
  out (cell v, cell e, cell volo, cell elemMass, cell calc_volume,
       cell p, cell q, cell sound_speed) @ -8.0{
  ℝ³ X[8];
  const ℝ chaos = (((ℝ)uid)+1.0)*option_chaos_seed;
  v=1.0;
  ∀ node X[n]=coord;
  e=(option_chaos)?chaos:(uid==0)?option_initial_energy:0.0;
  sound_speed=p=q=(option_chaos)?chaos;
  volo=elemMass=calc_volume=computeElemVolume(X);
  //debug()<<": e["<<uid<<"]="<<e<<", volume="<<calc_volume;
  //printf("\ne[%02d]=%.21e, volume=%.21e",uid,e,calc_volume);
}

∀ nodes void iniNodalMass(void) in (cell calc_volume) out (node nodalMass) @ -6.9{
  //const int nid=uid;
  nodalMass=0.0;
  //if (nid==0) printf("\n[iniNodalMass] n=%d, Node #%d, nodalMass=%.21e",n,nid,nodalMass);
  ∀ cell{
    /*
      const int a=xs_node_cell[8*nw+c];
    const int b=xs_node_cell[8*(nw+1)+c];
    const int xsa=2*WARP_BASE(a)+WARP_OFFSET(a);
    const int xsb=2*WARP_BASE(b)+WARP_OFFSET(b);
    debug()<<"a="<<a<<", xsa="<<xsa<<", b="<<b<<", xsb="<<xsb;
    */

    //int cid=node_cell[8*n+i];
    //if (cid<0) continue;
    //if (nid==0) printf("\n\tprefix: %d+=#%d nodalMass=%.21e (+=%.21e)",nid,cid,nodalMass,calc_volume/8.0);
    nodalMass+=calc_volume/8.0;
    //if (nid==0) printf("\n\tpostfix: %d+=#%d nodalMass=%.21e (+=%.21e)",nid,cid,nodalMass,calc_volume/8.0);
  }
  //debug()<<uid<<": nodalMass="<<nodalMass;
  //printf("\n#%d nodalMass=%.21e",uid,nodalMass);
}


// ****************************************************************************
// * Compute part @ ]+0,+∞[
// ****************************************************************************



// ****************************************************************************
// * timeIncrement
// * This routine computes the time increment δtn for the
// * current timestep loop iteration. We aim for a "target" value of t_final-tn
// * for δtn . However, the actual time increment is allowed to grow by a
// * certain prescribed amount from the value used in the previous step and is
// * subject to the constraints δt_Courant and δt_hydro described in Section 1.5.3.
// ****************************************************************************
void timeIncrement(void) @ 0.1 {
  const ℝ target_δt = option_stoptime - time;
  const ℝ max_δt = 1.0e+20;
  const ℝ new_δt_courant = (δt_courant < max_δt)?½*δt_courant:max_δt;
  const ℝ new_δt_courant_hydro = (δt_hydro < new_δt_courant)?δt_hydro*2.0/3.0:new_δt_courant;
  const ℝ now_δt = new_δt_courant_hydro ;
  const ℝ old_δt = (iteration==1)?option_δt_initial:δt;
  const ℝ ratio = now_δt / old_δt ;
  const ℝ up_new_δt = (ratio >= 1.0)?(ratio < option_δt_mult_lower_b)?old_δt:now_δt:now_δt;//option_δt_mult_lower_b
  const ℝ dw_new_δt = (ratio >= 1.0)?(ratio > option_δt_mult_upper_b)?old_δt*option_δt_mult_upper_b:up_new_δt:up_new_δt;
  const ℝ new_δt = (dw_new_δt > option_dtmax)?option_dtmax:dw_new_δt;
  const ℝ δτ = (option_δt_fixed <= 0.0)?(iteration != 1)?new_δt:old_δt:old_δt;
  const ℝ scaled_target_δt = (target_δt>δτ)?((target_δt<(4.0*δτ/3.0))?2.0*δτ/3.0:target_δt):target_δt;
  const ℝ scaled_δt = (scaled_target_δt < δτ)?scaled_target_δt:δτ;
  δt = scaled_δt ;
  //δt = (iteration>1)?1.0-07;
  //δt = (iteration==2)?1.199999999999999892758e-07;
  //δt = (iteration==3)?1.439999999999999924249e-07;
  //δt <?= (scaled_δt);
  //info()<<"ITERATION "<<iteration;
  //printf("\nITERATION %d, time=%f, δt=%f (sum=%f)",iteration,time,δt,time+δt);
  if (((time+δt)==option_stoptime)?true:false)
    assert(X_EDGE_ELEMS==8 && Y_EDGE_ELEMS==8 && Z_EDGE_ELEMS==8 && iteration==470);
  if (iteration >= option_max_iterations) exit;
  //info()<< "\t\t\33[7mTimeIncrement δt="<<δt<<"\33[m";//0]
  //printf("\n\t\t\33[7mTimeIncrement dtt=%.21e\33[m", δt);
  //if (δt == 0.) exit;
}


// ****************************************************************************
// * Sum contributions to total stress tensor
// * pull in the stresses appropriate to the hydro integration
// * Initialize stress terms for each element. Recall that our assumption of
// * an inviscid isotropic stress tensor implies that the three principal
// * stress components are equal, and the shear stresses are zero.
// * Thus, we initialize the diagonal terms of the stress tensor to
// * -(p + q) in each element.
// ****************************************************************************
∀ cells void initStressTermsForElems(void)
  in (node coord, cell p, cell q)
  out (cell ε, cell cForce) @ 0.3 {
  const ℝ chaos = (((ℝ)0.0)+1.0)*option_chaos_seed;
  const ℝ sig = (option_chaos)?chaos:-p-q;
  ℝ³ fNormals,dj,x[8],B[8];
  //debug()<<"sig="<<sig;
  ∀ node x[n] = coord;
  ε = dj = -¼*((x[0]+x[1]+x[5]+x[4])-(x[3]+x[2]+x[6]+x[7]));
  //calcElemShapeFunctionDerivatives(x,B);
  ∀ node B[n]=0.0;
  Σ_FaceNormal(B,0,1,2,3,x);
  Σ_FaceNormal(B,0,4,5,1,x);
  Σ_FaceNormal(B,1,5,6,2,x);
  Σ_FaceNormal(B,2,6,7,3,x);
  Σ_FaceNormal(B,3,7,4,0,x);
  Σ_FaceNormal(B,4,7,6,5,x);
  ∀ node cForce = -sig*B[n];
  //printf("\n@0.3 cForce[%d]=(%.21e,%.21e,%.21e)",uid,cForce.x,cForce.y,cForce.z);
  }

∀ nodes void fetchCellNodeForce0(void) in (cell cForce) out (node nForce)@0.301 {
  ℝ³ Σ=ℝ³(0.0,0.0,0.0);
  ∀ cell{
    /*
      const int a=xs_node_cell[8*nw+c];
    const int b=xs_node_cell[8*(nw+1)+c];
    const int xsa=2*WARP_BASE(a)+WARP_OFFSET(a);
    const int xsb=2*WARP_BASE(b)+WARP_OFFSET(b);
    debug()<<"a="<<a<<", xsa="<<xsa<<", b="<<b<<", xsb="<<xsb;
    */
    //const int corner = node_cell_corner[8*n+i];
    //if (corner==-1) continue;
    //printf("\n[#%d] node_#%d is corner %d of cell %d",i,uid,corner,node_cell[8*n+i]);
    Σ+=cForce;
  }
  nForce=Σ;
  //debug()<<"@0.3 nForce="<<nForce;
  //printf("\n@0.3 nForce[%d]=(%.21e,%.21e,%.21e)",uid,nForce.x,nForce.y,nForce.z);
}

// ****************************************************************************
// * calcFBHourglassForceForElems
// * Calculates the Flanagan-Belytschko anti-hourglass force
// * calcFBHourglassForceForElems
// ****************************************************************************
∀ cells void ΣElemStressesToNodeForces(void)
  in (node coord, cell volo, cell v,
      cell sound_speed, cell elemMass,
      node 𝜕x)
  out (cell cForce)  @ 1.3{
  const ℝ γ[4][8]={{ 1., 1.,-1.,-1.,-1.,-1., 1., 1.},
                      { 1.,-1.,-1., 1.,-1., 1., 1.,-1.},
                      { 1.,-1., 1.,-1., 1.,-1., 1.,-1.},
                      {-1., 1.,-1., 1., 1.,-1., 1.,-1.}};
  ℝ η0[4],η1[4],η2[4],η3[4] ;
  ℝ η4[4],η5[4],η6[4],η7[4];
  ℝ³ x[8],xd[8],dvd[8],η[8];
  const ℝ hourg=option_hgcoef;
  const ℝ τv = volo*v;
  const ℝ volume13=∛(τv);
  const ℝ θ = -hourg*0.01*sound_speed*elemMass/volume13;
  const ℝ determ = τv;
  //const ℝ chaos = (((ℝ)uid)+1.0)*option_chaos_seed;
  ∀ node x[n] = coord;
  //∀ node xd[n] = (option_chaos)?chaos:𝜕x;  
  ∀ node xd[n] = 𝜕x;  
  dvd[0]=𝜕Volume(x[1],x[2],x[3],x[4],x[5],x[7]);
  dvd[3]=𝜕Volume(x[0],x[1],x[2],x[7],x[4],x[6]);
  dvd[2]=𝜕Volume(x[3],x[0],x[1],x[6],x[7],x[5]);
  dvd[1]=𝜕Volume(x[2],x[3],x[0],x[5],x[6],x[4]);
  dvd[4]=𝜕Volume(x[7],x[6],x[5],x[0],x[3],x[1]);
  dvd[5]=𝜕Volume(x[4],x[7],x[6],x[1],x[0],x[2]);
  dvd[6]=𝜕Volume(x[5],x[4],x[7],x[2],x[1],x[3]);
  dvd[7]=𝜕Volume(x[6],x[5],x[4],x[3],x[2],x[0]);
  cHourglassModes(0,determ,dvd,γ,x,η0,η1,η2,η3,η4,η5,η6,η7);
  //printf("\n[#%d] η[0]: %.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e",uid, η0[0],η1[0],η2[0],η3[0],η4[0],η5[0],η6[0],η7[0]);
  cHourglassModes(1,determ,dvd,γ,x,η0,η1,η2,η3,η4,η5,η6,η7);
  //rintf("\n[#%d] η[1]: %.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e",uid, η0[1],η1[1],η2[1],η3[1],η4[1],η5[1],η6[1],η7[1]);
  cHourglassModes(2,determ,dvd,γ,x,η0,η1,η2,η3,η4,η5,η6,η7);
  //printf("\n[#%d] η[2]: %.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e",uid, η0[2],η1[2],η2[2],η3[2],η4[2],η5[2],η6[2],η7[2]);
  cHourglassModes(3,determ,dvd,γ,x,η0,η1,η2,η3,η4,η5,η6,η7);
  //printf("\n[#%d] η[3]: %.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e,%.21e",uid, η0[3],η1[3],η2[3],η3[3],η4[3],η5[3],η6[3],η7[3]);
  //debug()<<"hourg="<<hourg<<", sound_speed="<<sound_speed<<", elemMass="<<elemMass;
  //printf("\nθ[%d]=%.21e",uid,θ);
  calcElemFBHourglassForce(xd,η0,η1,η2,η3,η4,η5,η6,η7,θ,η);
  //foreach node debug()<<uid<<" "<<n<<" hgf="<<η[n];
  //∀ node printf("\n[#%d] η[%d]=(%.21e,%.21e,%.21e)",uid,n,η[n].x,η[n].y,η[n].z);
  ∀ node cForce = η[n];
//#warning cForce chaos
    //foreach node cForce += η[n];
  //foreach node cForce = (n+1.0)*1.1234567890123456789;
  //printf("\n@1.3 cForce[%d]=(%.21e,%.21e,%.21e)",uid,cForce.x,cForce.y,cForce.z);
}

∀ nodes void ΣCellNodeForce(void) in (cell cForce) out (node nForce)@ 1.4 {
  ℝ³ Σ=ℝ³(0.0,0.0,0.0);//0.0;
  foreach cell Σ+=cForce;
  nForce+=Σ;
  //debug()<<"@1.4 nForce="<<nForce;
  //printf("\n@1.4 nForce[%d]=(%.21e,%.21e,%.21e)",uid,nForce.x,nForce.y,nForce.z);
}


// ****************************************************************************
// * The routine CalcAccelerationForNodes() calculates a three-dimensional
// * acceleration vector A at each mesh node from F.
// * The acceleration is computed using Newton's Second Law of Motion,
// * F = m0 A, where m0 is the mass at the node.
// * Note that since the mass in each element is constant in time for our calculations,
// * the mass at each node is also constant in time.
// * The nodal mass values are set during the problem set up.
// ****************************************************************************
∀ nodes void 𝜕𝜕ForNodes(void)
  in (node nForce, node nodalMass) out (node 𝜕𝜕x) @ 2.8{
  𝜕𝜕x = nForce/nodalMass;
  //debug()<<"𝜕𝜕x="<<𝜕𝜕x;
}

// ****************************************************************************
// * The routine ApplyAccelerationBoundaryConditions() applies symmetry boundary
// * conditions at nodes on the boundaries of the mesh where these were specified
// * during problem set up. A symmetry boundary condition sets the normal
// * component of A at the boundary to zero.
// * This implies that the normal component of the velocity vector U will
// * remain constant in time.
// ****************************************************************************
∀ outer nodes void 𝜕𝜕BCForNodes(void)
  out (node 𝜕𝜕x) @ 2.9 {
  𝜕𝜕x.x=(coord.x==0.0)?0.0;
  𝜕𝜕x.y=(coord.y==0.0)?0.0;
  𝜕𝜕x.z=(coord.z==0.0)?0.0;
}


// ****************************************************************************
// * The routine CalcVelocityForNodes() integrates the acceleration at each node
// * to advance the velocity at the node to tn+1.
// * Note that this routine also applies a cut-off to each velocity vector value.
// * Specifically, if a value is below some prescribed value, that term is set to zero.
// * The reason for this cutoff is to prevent spurious mesh motion which may arise
// * due to floating point roundoff error when the velocity is near zero.
// ****************************************************************************
∀ nodes void 𝜕ForNodes(void) in (node 𝜕𝜕x) inout (node 𝜕x) @ 3.0{
  𝜕x += 𝜕𝜕x*δt ;
  𝜕x.x = (norm(𝜕x.x)<option_u_cut)?0.0;
  𝜕x.y = (norm(𝜕x.y)<option_u_cut)?0.0;
  𝜕x.z = (norm(𝜕x.z)<option_u_cut)?0.0;
  //debug()<<"𝜕x="<<𝜕x;
  //printf("\nvelocity[%d]=(%.21e,%.21e,%.21e)",uid,𝜕x.x,𝜕x.y,𝜕x.z);
}


// ****************************************************************************
// * The routine CalcPositionForNodes() performs the last step in the nodal
// * advance portion of the algorithm by integrating the velocity at each node
// * to advance the position of the node to tn+1.
// ****************************************************************************
∀ nodes void coordForNodes(void) in (node 𝜕x) @ 3.1{
  coord += 𝜕x*δt;
  //debug()<<"coords= "<<coord;
  //printf("\ncoords[%d]=(%.21e,%.21e,%.21e)",uid,coord.x,coord.y,coord.z);
}


// ****************************************************************************
// * calcElemVolume
// ****************************************************************************
∀ cells void calcElemVolume(void)
  in (node coord, node 𝜕x,cell v, volo, delx_xi, delv_xi,
      cell delx_eta, delv_eta, delx_zeta, delv_zeta)
  inout (cell ε) out (cell delv, vnew, calc_volume, vdov, arealg) @ 4.0{
  const ℝ dt2= ½*δt;
  const ℝ δ = 1.e-36;
  ℝ³ B[8],X[8],Xd[8];
  ℝ DetJ,volume,ρVolume;
  foreach node X[n]=coord;
  foreach node Xd[n]=𝜕x;
  //printf("\n[#%d] Xd[1]=(%.21e,%.21e,%.21e)",uid, Xd[1].x, Xd[1].y, Xd[1].z);
  volume = calc_volume = computeElemVolume(X);  
  vnew = ρVolume = volume/volo;
  delv = ρVolume - v;
  //printf("\n[#%d] volume=%.21e, volo=%.21e, ρVolume=%.21e, delv=%.21e",uid, volume, volo, ρVolume, delv);
  arealg = calcElemCharacteristicLength(X,volume);
  //debug()<<"arealg="<<arealg;
  {
    const ℝ vol = volo*vnew;
    const ℝ nrm = 1.0/(vol+δ);
    const ℝ³ di =  ¼*((X[1]+X[2]+X[6]+X[5])-(X[0]+X[3]+X[7]+X[4]));
    const ℝ³ dj = -¼*((X[0]+X[1]+X[5]+X[4])-(X[3]+X[2]+X[6]+X[7]));
    const ℝ³ dk =  ¼*((X[4]+X[5]+X[6]+X[7])-(X[0]+X[1]+X[2]+X[3]));
    const ℝ³ a_xi = (dj⨯dk);
    const ℝ³ a_eta = (dk⨯di);
    const ℝ³ a_zeta = (di⨯dj);
    const ℝ³ dv_xi  =  ¼*((Xd[1]+Xd[2]+Xd[6]+Xd[5])-(Xd[0]+Xd[3]+Xd[7]+Xd[4]));
    const ℝ³ dv_eta = -¼*((Xd[0]+Xd[1]+Xd[5]+Xd[4])-(Xd[3]+Xd[2]+Xd[6]+Xd[7]));
    const ℝ³ dv_zeta = ¼*((Xd[4]+Xd[5]+Xd[6]+Xd[7])-(Xd[0]+Xd[1]+Xd[2]+Xd[3]));
    delx_xi = vol/√(a_xi⋅a_xi+δ);
    delx_eta = vol/√(a_eta⋅a_eta+δ);
    delx_zeta = vol/√(a_zeta⋅a_zeta+δ);
    delv_zeta = (a_zeta*nrm)⋅dv_zeta;     
    delv_xi = (a_xi*nrm)⋅dv_xi;
    delv_eta = (a_eta*nrm)⋅dv_eta;
  }
  foreach node X[n] -= dt2*Xd[n];
  DetJ=calcElemShapeFunctionDerivatives(X,B);
  //printf("\n[#%d] DetJ=%.21e",uid, DetJ);
  //printf("\n[#%d] B[0]=(%.21e,%.21e,%.21e)",uid, B[0].x, B[0].y, B[0].z);
  //printf("\n[#%d] B[1]=(%.21e,%.21e,%.21e)",uid, B[1].x, B[1].y, B[1].z);
  //printf("\n[#%d] B[2]=(%.21e,%.21e,%.21e)",uid, B[2].x, B[2].y, B[2].z);
  //printf("\n[#%d] B[3]=(%.21e,%.21e,%.21e)",uid, B[3].x, B[3].y, B[3].z);
  
  //printf("\n[#%d] Xd[0]=(%.21e,%.21e,%.21e)",uid, Xd[0].x, Xd[0].y, Xd[0].z);
  //printf("\n[#%d] Xd[1]=(%.21e,%.21e,%.21e)",uid, Xd[1].x, Xd[1].y, Xd[1].z);
  //printf("\n[#%d] Xd[2]=(%.21e,%.21e,%.21e)",uid, Xd[2].x, Xd[2].y, Xd[2].z);
  //printf("\n[#%d] Xd[3]=(%.21e,%.21e,%.21e)",uid, Xd[3].x, Xd[3].y, Xd[3].z);
  //printf("\n[#%d] Xd[4]=(%.21e,%.21e,%.21e)",uid, Xd[4].x, Xd[4].y, Xd[4].z);
  //printf("\n[#%d] Xd[5]=(%.21e,%.21e,%.21e)",uid, Xd[5].x, Xd[5].y, Xd[5].z);
  //printf("\n[#%d] Xd[6]=(%.21e,%.21e,%.21e)",uid, Xd[6].x, Xd[6].y, Xd[6].z);
  //printf("\n[#%d] Xd[7]=(%.21e,%.21e,%.21e)",uid, Xd[7].x, Xd[7].y, Xd[7].z);
  ε=calcElemVelocityGradient(Xd,B,DetJ);
  //printf("\n[#%d] Epsilon=(%.21e,%.21e,%.21e)",uid, ε.x, ε.y, ε.z);
  //calc strain rate and apply as constraint
  vdov = ε.x+ε.y+ε.z;
  //printf("\n[#%d] vdov=%.21e",uid, vdov);
  // make the rate of deformation tensor deviatoric
  //ε -= ⅓*vdov;
  const ℝ tvdov = ⅓*vdov;
  const ℝ³ t3vdov = ℝ³(tvdov,tvdov,tvdov);
  ε -= t3vdov;
  // See if any volumes are negative, and take appropriate action.
  //assert(vnew >= 0.0);//fatal("calcLagrangeElements","VolumeError");
}


// ****************************************************************************
// * This routine performs the second part of the q calculation.
// ****************************************************************************
∀ cells void calcMonotonicQForElemsByDirectionX(void)
  in (cell elemBC, cell delv_xi) out (cell phixi) @ 4.7 {
  const ℝ monoq_limiter_mult = option_monoq_limiter_mult;
  const ℝ monoq_max_slope = option_monoq_max_slope;
  ℤ bcSwitch;
  ℝ delvm=0.0;
  ℝ delvp=0.0;
  const ℝ ptiny = 1.e-36;
  const ℝ nrm = 1./(delv_xi+ptiny);
  bcSwitch = elemBC & XI_M;
  delvm = (bcSwitch == 0)?delv_xi[prevCellX];
  delvm = (bcSwitch == XI_M_SYMM)?delv_xi;
  delvm = (bcSwitch == XI_M_FREE)?0.0;
  delvm = delvm * nrm ;
  bcSwitch = elemBC & XI_P;
  delvp = (bcSwitch == 0)?delv_xi[nextCellX];
  delvp = (bcSwitch == XI_P_SYMM)?delv_xi;
  delvp = (bcSwitch == XI_P_FREE)?0.0;
  delvp = delvp * nrm ;
  phixi = ½ * (delvm + delvp) ;
  delvm *= monoq_limiter_mult ;
  delvp *= monoq_limiter_mult ;
  phixi = (delvm < phixi)?delvm;
  phixi = (delvp < phixi)?delvp;
  phixi = (phixi < 0.)?0.0;
  phixi = (phixi > monoq_max_slope)?monoq_max_slope;
  }

∀ cells void calcMonotonicQForElemsByDirectionY(void)
  in (cell elemBC, cell delv_eta) out (cell phieta) @ 4.7 {
  const ℝ monoq_limiter_mult = option_monoq_limiter_mult;
  const ℝ monoq_max_slope = option_monoq_max_slope;
  ℤ bcSwitch;
  ℝ delvm=0.;
  ℝ delvp=0.;
  const ℝ ptiny = 1.e-36;
  const ℝ nrm = 1./(delv_eta+ptiny);
  bcSwitch = elemBC & ETA_M;
  delvm = (bcSwitch == 0)?delv_eta[prevCellY];
  delvm = (bcSwitch == ETA_M_SYMM)?delv_eta;
  delvm = (bcSwitch == ETA_M_FREE)?0.0;
  delvm = delvm * nrm ;
  bcSwitch = elemBC & ETA_P;
  delvp = (bcSwitch == 0)?delv_eta[nextCellY];
  delvp = (bcSwitch == ETA_P_SYMM)?delv_eta;
  delvp = (bcSwitch == ETA_P_FREE)?0.0;
  delvp = delvp * nrm ;
  phieta = ½*(delvm + delvp) ;
  delvm *= monoq_limiter_mult ;
  delvp *= monoq_limiter_mult ;
  phieta = (delvm  < phieta)?delvm;
  phieta = (delvp  < phieta)?delvp;
  phieta = (phieta < 0.0)?0.0;
  phieta = (phieta > monoq_max_slope)?monoq_max_slope;
}

∀ cells void calcMonotonicQForElemsByDirectionZ(void)
  in (cell elemBC, cell delv_zeta) out (cell phizeta) @ 4.7 {
  const ℝ monoq_limiter_mult = option_monoq_limiter_mult;
  const ℝ monoq_max_slope = option_monoq_max_slope;
  ℤ bcSwitch;
  ℝ delvm=0.;
  ℝ delvp=0.;
  const ℝ ptiny = 1.e-36;
  const ℝ nrm = 1./(delv_zeta+ptiny) ;
  bcSwitch = elemBC & ZETA_M;
  delvm = (bcSwitch == 0)?delv_zeta[prevCellZ];
  delvm = (bcSwitch == ZETA_M_SYMM)?delv_zeta;
  delvm = (bcSwitch == ZETA_M_FREE)?0.0;
  delvm = delvm * nrm ;
  bcSwitch = elemBC & ZETA_P;
  delvp = (bcSwitch == 0)?delv_zeta[nextCellZ];
  delvp = (bcSwitch == ZETA_P_SYMM)?delv_zeta;
  delvp = (bcSwitch == ZETA_P_FREE)?0.0;
  delvp = delvp * nrm ;
  phizeta = ½*(delvm+delvp);
  delvm *= monoq_limiter_mult ;
  delvp *= monoq_limiter_mult ;
  phizeta = (delvm < phizeta )?delvm;
  phizeta = (delvp < phizeta )?delvp;
  phizeta = (phizeta < 0.0)?0.0;
  phizeta = (phizeta > monoq_max_slope )?monoq_max_slope;
}

∀ cells void calcMonotonicQForElemsQQQL(void)
  in (cell vdov, cell elemMass, cell volo, cell vnew,
      cell delx_xi, cell delv_eta, cell delx_eta,
      cell delv_zeta, cell delx_zeta, cell delv_xi,
      cell phixi, cell phieta, cell phizeta)
  out(cell qq, cell ql)@ 4.72{
  const ℝ rho = elemMass/(volo*vnew);
  const ℝ qlc_monoq = option_qlc_monoq;
  const ℝ qqc_monoq = option_qqc_monoq;
  const ℝ delvxxi   = delv_xi*delx_xi;
  const ℝ delvxeta  = delv_eta*delx_eta;
  const ℝ delvxzeta = delv_zeta*delx_zeta;
  const ℝ delvxxit  = (delvxxi>0.0)?0.0:delvxxi;
  const ℝ delvxetat = (delvxeta>0.0)?0.0:delvxeta;
  const ℝ delvxzetat= (delvxzeta>0.0)?0.0:delvxzeta;
  const ℝ qlin = -qlc_monoq*rho*(delvxxit*(1.0-phixi)+
                                    delvxetat*(1.0-phieta)+
                                    delvxzetat*(1.0-phizeta));
  const ℝ qquad = qqc_monoq*rho*(delvxxit*delvxxit*(1.0-phixi*phixi) +
                                    delvxetat*delvxetat*(1.0-phieta*phieta) +
                                    delvxzetat*delvxzetat*(1.0-phizeta*phizeta));
  const ℝ qlint  = (vdov>0.0)?0.0:qlin;  // Remove length scale
  const ℝ qquadt = (vdov>0.0)?0.0:qquad;
  qq = qquadt ;
  ql = qlint  ;
  //debug()<<"phixi="<<phixi<<" phieta="<<phieta<<" phizeta="<<phizeta;
  //printf("\nphixi=%.21e phieta=%.21e phizeta=%.21e",phixi,phieta,phizeta);
  //debug()<<"ql="<<ql<<", qq="<<qq;
  //ok printf("\n[#%02d] ql=%.21e, qq=%.21e",uid, ql,qq);
}



// ****************************************************************************
// * The routine ApplyMaterialPropertiesForElems() updates the pressure and
// * internal energy variables to their values at the new time, p_n+1 and e_n+1.
// * The routine first initializes a temporary array with the values of Vn+1 for
// * each element that was computed earlier. Then, upper and lower cut-off
// * values are applied to each array value to keep the relative volumes
// * within a prescribed range (not too close to zero and not too large).
// * Next, the routine EvalEOSForElems() is called and the array of modified
// * relative element volumes is passed to it.
// ****************************************************************************
∀ cells void applyMaterialPropertiesForElems(void)
  in (cell vnew) out(cell vnewc) @ 5.0{
  vnewc = vnew ;
  vnewc = (vnewc < option_eosvmin)?option_eosvmin;
  vnewc = (vnewc > option_eosvmax)?option_eosvmax;
}


// ****************************************************************************
// * The routine EvalEOSForElems() calculates updated values for pressure p_n+1
// * and internal energy e_n+1.
// * The computation involves several loops over elements to pack various mesh
// * element arrays (e.g., p, e, q, etc.) into local temporary arrays.
// * Several other quantities are computed and stored in element length
// * temporary arrays also.
// * The temporary arrays are needed because the routine CalcEnergyForElems()
// * calculates p_n+1 and e_n+1 in each element in an iterative process that
// * requires knowledge of those variables at time tn while it computes the
// * new time values.
// ****************************************************************************
∀ cells void evalEOSForElems0(void)
  in (cell e, cell delv, cell p, cell q, cell vnewc)
  out(cell e_old, cell delvc, cell p_old, cell q_old,
      cell compression, cell compHalfStep,cell work) @ 6.0{
  const ℝ vchalf = vnewc - ( ½*delv);
  work = 0.0; 
  e_old = e;
  delvc = delv;
  p_old = p;
  q_old = q ;
  compression = (1.0/vnewc) - 1.0;
  compHalfStep = (1.0/vchalf)-1.0;
}

∀ cells void evalEOSForElems1(void)
  in (cell vnewc, cell compression)
  out(cell compHalfStep) @ 6.1 {
  compHalfStep = (vnewc <= option_eosvmin)?compression;
}

∀ cells void evalEOSForElems6(void)
  in (cell vnewc, cell compHalfStep)
  out(cell p_old, cell compression) @ 6.6 {
  p_old = (vnewc < option_eosvmax)?p_old:0.0;
  compression =(vnewc < option_eosvmax)?compression:0.0;
  compHalfStep = (vnewc < option_eosvmax)?compHalfStep:0.0;
}


// ****************************************************************************
// * The routine CalcEnergyForElems() calls CalcPressureForElems() repeatedly.
// * The function CalcPressureForElems() is the Equation of State model
// * for a "gamma law" gas.
// * The value c1s passed to the routine is defined to be γ - 1.
// * The Equation of State calculation is a core part of any hydrocode.
// * In a production code, one of any number of Equation of State functions
// * may be called to generate a pressure that is needed to close the system
// * of equations and generate a unique solution.
// ****************************************************************************
// @ 7.1   calcEnergyForElems1
∀ cells void calcEnergyForElems1(void)
  in (cell e_old, cell delvc, cell p_old, cell q_old, cell work)
  inout (cell e_new) @ 7.1{
  e_new = e_old - ½*delvc*(p_old + q_old) + ½*work;
  e_new = (e_new < option_emin)?option_emin;
  //debug()<<"vchalf="<<(vnewc-(½*delv));
  //debug()<<"e_old="<<e_old<<", delvc="<< delvc <<", work="<< work<<", e_new="<<e_new;
  //ok printf("\n[#%d] e_old=%.21e, delvc=%.21e, work=%.21e, e_new=%.21e",uid,e_old, delvc, work, e_new);
}


// ****************************************************************************
// * calcPressureForElems
// * p_new => pHalfStep
// * compression => compHalfStep
// * e_old => e_new
// ****************************************************************************
∀ cells void calcPressureForElemspHalfStepcompHalfStep(void)
  in (cell compHalfStep, cell e_new) 
  inout(cell bvc, cell pHalfStep)
  out (cell vnewc, cell pbvc) @ 7.2{
  const ℝ c1s = 2.0/3.0;
  bvc = c1s*(compHalfStep+1.0);
  pbvc = c1s;
  pHalfStep = bvc*e_new ;
  pHalfStep=(rabs(pHalfStep)<option_p_cut)?0.0;
  pHalfStep = (vnewc >= option_eosvmax )?0.0; // impossible condition here?
  pHalfStep = (pHalfStep < option_pmin)?option_pmin;
  //debug()<<"pHalfStep="<<pHalfStep<<", bvc="<<bvc<<", pbvc="<<pbvc<<", e_new="<<e_new;
  //printf("\npHalfStep=%.21e, bvc=%.21e, pbvc=%.21e, e_new=%.21e",pHalfStep, bvc, pbvc, e_new);
  //debug()<<"compHalfStep="<<compHalfStep<<", vnewc="<<vnewc<<", pmin="<<option_pmin<<", p_cut="<<option_p_cut;
}

inline ℝ computeSoundSpeed(const ℝ c,//pbvc
                           const ℝ energy,
                           const ℝ volume,
                           const ℝ b,//bvc
                           const ℝ pressure,
                           const ℝ rho,
                           const ℝ _ql,
                           const ℝ _qq){
  const ℝ pe = c*energy;
  const ℝ vvbp=volume*volume*b*pressure;
  const ℝ ssc = (pe + vvbp)/rho;
  const ℝ ssct = (ssc <= 0.111111e-36)?0.333333e-18:√(ssc);
  const ℝ sscq = ssct*_ql;
  const ℝ sscqt = sscq+_qq;
  return sscqt;
}
inline ℝ computeSoundSpeed(const ℝ c,//pbvc
                           const ℝ energy,
                           const ℝ volume,
                           const ℝ b,//bvc
                           const ℝ pressure,
                           const ℝ rho){
  const ℝ pe = c*energy;
  const ℝ vvbp=volume*volume*b*pressure;
  const ℝ ssc = (pe + vvbp)/rho;
  const ℝ ssct = (ssc <= 0.111111e-36)?0.333333e-18:√(ssc);
  return ssct;
}

∀ cells void calcEnergyForElems3(void)
  in (cell compHalfStep, cell delvc, cell pbvc, cell ql, cell qq,
      cell bvc, cell pHalfStep, cell p_old, cell q_old)
  out (cell q_new)
  inout (cell e_new) @ 7.3 {
  const ℝ vhalf = 1.0/(1.0+compHalfStep);
  const ℝ ssc = computeSoundSpeed(pbvc,e_new,vhalf,bvc,pHalfStep,option_refdens,ql,qq);
  q_new = (delvc>0.0)?0.0:ssc;//(ssct*ql + qq);
  e_new = e_new + ½*delvc*(3.0*(p_old+q_old)-4.0*(pHalfStep+q_new)) ;
  //ok printf("\n[#%02d] q_new=%.21e, e_new=%.21e",uid,q_new,e_new);
}

∀ cells void calcEnergyForElems4(void) in (cell work)
  inout (cell e_new)@ 7.4{
  e_new += ½*work;
  e_new = (rabs(e_new) < option_e_cut)?0.0;
  e_new = (e_new<option_emin)?option_emin;
  //debug()<<"q_new="<<q_new<<", e_new="<<e_new;
  //printf("\n[#%02d] e_new=%.21e",uid,e_new);
}

∀ cells void calcPressureForElemsPNewCompression(void)
  in (cell compression,
      cell bvc,
      cell e_new, cell vnewc)
      inout (cell pbvc, cell p_new) @ 7.5,7.7{
  const ℝ c1s = 2.0/3.0;
  bvc = c1s*(compression + 1.0);
  pbvc = c1s;
  p_new = bvc*e_new ;
  p_new = (rabs(p_new) < option_p_cut)?0.0;
  p_new = (vnewc >= option_eosvmax )?0.0;
  p_new = (p_new < option_pmin)?option_pmin;
}

∀ cells void calcEnergyForElems6(void)
  in (cell delvc, cell bvc,cell pbvc, cell vnewc, cell p_new, cell ql,
      cell qq, cell p_old, cell q_old, cell pHalfStep, cell q_new)
  inout(cell e_new)@ 7.6{
  const ℝ sixth = 1.0/6.0;
  //const ℝ ssc = ( pbvc*e_new + vnewc*vnewc*bvc*p_new)/option_refdens;
  //const ℝ ssct = (ssc <= 0.111111e-36)?0.333333e-18:√(ssc);
  const ℝ ssc = computeSoundSpeed(pbvc,e_new,vnewc,bvc,p_new,option_refdens,ql,qq);
  const ℝ q_tilde = (delvc > 0.)?0.0:ssc;//(ssct*ql + qq);
  //debug()<<"ssc="<<ssct;
  //debug()<<"pbvc="<<pbvc<<" e_new="<<e_new<<" vnewc="<<vnewc
  //<<" bvc="<<bvc<<" p_new="<<p_new <<" rho0="<<option_refdens;
  e_new = e_new - (7.0*(p_old + q_old)
                   - (8.0)*(pHalfStep + q_new)
                   + (p_new + q_tilde)) * delvc*sixth ;
  //debug()<<"e_new="<<e_new;
  e_new = (rabs(e_new) < option_e_cut)?0.0;
  e_new = (e_new < option_emin)?option_emin;
  //debug()<<"e_new="<<e_new;
  //ok printf("\n[#%02d] e_new=%.21e",uid,e_new);
}

∀ cells void calcEnergyForElems8(void)
  in (cell delvc, cell bvc, cell pbvc, cell e_new,
      cell vnewc, cell p_new, cell ql, cell qq)
  inout(cell q_new)@ 7.8{
  const ℝ qnw = computeSoundSpeed(pbvc,e_new,vnewc,bvc,p_new,option_refdens,ql,qq);
  const ℝ qnwt = (rabs(qnw) < option_q_cut)?0.0:qnw;
  //printf("\n[#%02d] pe=%.21e, vvbp=%.21e, ss=%.21e",uid,pe,vvbp,ss);
  //printf("\n[#%02d] ssc=%.21e, ssct=%.21e, sq=%.21e, qnw=%.21e, qnwt=%.21e",uid,ssc,ssct,sq,qnw,qnwt);
  //debug()<<"pbvc="<<pbvc<<" e_new="<<e_new<<" vnewc="<<vnewc
  //<<" bvc="<<bvc<<" p_new="<<p_new <<" rho0="<<option_refdens;
  //debug()<<"8 delvc="<<delvc;
  //debug()<<"8 ssc="<<ssct;
  q_new = (delvc <= 0.)?qnwt;
  //debug()<<"8 q_new="<<q_new;
  //printf("\n[#%02d] pe=%.21e, vvbp=%.21e, ssc=%.21e, ssct=%.21e, qnw=%.21e, qnwt=%.21e",uid,pe,vvbp,ssc,ssct,qnw,qnwt);
  //printf("\n[#%02d] q_new=%.21e",uid,q_new);
}

∀ cells void evalEOSForElems8(void)
  in (cell p_new, cell e_new, cell q_new)
  out(cell p, cell e, cell q)@ 8.0{
  p = p_new;
  e = e_new;
  q = q_new;
  //debug()<<"p_new="<<p_new<<" e_new="<< e_new <<" q_new="<< q_new;
  //printf("\n[#%02d] p_new=%.21e, e_new=%.21e, q_new=%.21e",uid,p_new, e_new, q_new);
}


// ****************************************************************************
// * Lastly, the routine CalcSoundSpeedForElems() calculates the sound speed
// * sound_speed in each element using p_n+1 and e_n+1.
// * The maximum value of sound_speed is used to calculate constraints on t_n+1
// * which will be used for the next time advance step.
// ****************************************************************************
∀ cells void calcSoundSpeedForElems(void)
  in (cell bvc, cell pbvc, cell e_new, cell vnewc, cell p_new)
  out (cell sound_speed) @ 9.0{
  //foreach material
  {
    //debug()<<"pbvc="<<pbvc<<" e_new="<<e_new<<" vnewc="<<vnewc<<" bvc="<<bvc<<" p_new="<<p_new <<" rho0="<<option_refdens;
    //const ℝ ssTmp = (pbvc*e_new + vnewc*vnewc*bvc*p_new)/option_refdens;
    //const ℝ ssTmpt = (ssTmp <= 0.111111e-36)?0.333333e-18:√(ssTmp);
    const ℝ ssTmpt = computeSoundSpeed(pbvc,e_new,vnewc,bvc,p_new,option_refdens);
    sound_speed = ssTmpt;
    //debug()<<"calcSoundSpeedForElems sound_speed="<<sound_speed;
    //ok printf("\n[#%d] sound_speed=%.21e",uid,sound_speed);
  }
}


// ****************************************************************************
// * The routine UpdateVolumesForElems() updates the relative volume to V_n+1.
// * This routine basically resets the current volume V_n in each element to
// * the new volume V_n+1 so the simulation can continue to the next time
// * increment.
// * Note that this routine applies a cut-off to the relative volume V in
// * each element. Specifically, if V is sufficiently close to one (a
// * prescribed tolerance), then V is set to one.
// * The reason for this cutoff is to prevent spurious deviations of volume
// * from their initial values which may arise due to floating point roundoff
// * error.
// ****************************************************************************
∀ cells void updateVolumesForElems(void)
  in (cell vnew) out (cell v) @ 10.0{
  const ℝ ν = vnew;
  const ℝ νt = (rabs(ν-1.0)<option_v_cut)?1.0:ν;
  v = νt;
  //debug()<<"updateVolumesForElems v="<<v;
  //ok printf("\n[#%d] v=%.21e",uid,v);
}


// ****************************************************************************
// * The routine CalcCourantConstraintForElems() calculates the Courant timestep
// * constraint δt_Courant. This constraint is calculated only in elements
// * whose volumes are changing that is, dV/V!=0.
// * If all element volumes remain the same, there is no Courant constraint
// * applied during the subsequent δt calculation.
// ****************************************************************************
∀ cells void calcCourantConstraintForElems(void)
  in (cell sound_speed, cell arealg, cell vdov)
  out (cell δt_cell_courant) @ 12.1{
  const ℝ arg_max_courant=1.0e+20;
  δt_cell_courant=arg_max_courant;
  //foreach material
  {
    const ℝ qqc2 = 64.0 * option_qqc * option_qqc ;
    const ℝ δf = sound_speed[m] * sound_speed[m];
    const ℝ δft=(vdov[m]<0.0)?qqc2*arealg[m]*arealg[m]*vdov[m]*vdov[m]:0.0;
    const ℝ δfpp = δf+δft;
    const ℝ δfp = √(δfpp);
    const ℝ aδfp = arealg[m]/δfp;
    //ok printf("\n[#%d] vdov=%.21e, arealg=%.21e, sound_speed=%.21e, aδfp=%.21e",uid,vdov,arealg,sound_speed,aδfp);
    δt_cell_courant=(vdov!=0.0)?min(arg_max_courant,aδfp);
    //debug()<<"dtcourant="<<dtcourant;
    //ok printf("\n[#%d] dtcourant=%.21e",uid,dtcourant);
  }
  //return dtcourant;
} 


// ****************************************************************************
// * The routine CalcHydroConstraintForElems() calculates the hydro timestep
// * constraint. Similar to δt_Courant, δt_hydro is calculated only in elements
// * whose volumes are changing. When an element is undergoing volume change,
// * δt_hydro for the element is some maximum allowable element volume change
// * (prescribed) divided by dV/V in the element.
// ****************************************************************************
∀ cells void calcHydroConstraintForElems(void)
  in (cell vdov) out (cell δt_cell_hydro) @ 12.2{
  const ℝ arg_max_hydro=1.0e+20;
  δt_cell_hydro=arg_max_hydro;
  //foreach material
  {
    const ℝ δdv = rabs(vdov[m]);
    const ℝ δdve = δdv+1.e-20;
    const ℝ δdvov = option_dvovmax/δdve;
    const ℝ δhdr = min(arg_max_hydro,δdvov);
    //debug()<<"vdov="<<vdov<<" dtdvov="<<dtdvov<<" dvovmax="<<option_dvovmax;
    //printf("\n[#%d] vdov=%.21e, dtdvov=%.21e, dvovmax=%.21e,",uid,vdov,dtdvov,option_dvovmax);
    δt_cell_hydro=(vdov!=0.0)?δhdr;
    //info()<<"δt_cell_hydro="<<δt_cell_hydro;
    //printf("\n[#%d] δdvov=%.21e, δhdr=%.21e, dthydro=%.21e",uid,δdvov,δhdr,dthydro);
    //printf("\n[#%d] δt_cell_hydro=%.21e",uid,δt_cell_hydro);
  }
  //return dthydro;
}


// ****************************************************************************
// * After all solution variables are advanced to t_n+1, the constraints δtCourant
// * and δthydro for the next time increment t_n+1 are calculated in this routine.
// * Each constraint is computed in each element and then the final constraint value
// * is the minimum over all element values.
// * The constraints are applied during the computation of δt for the next time step.
// ****************************************************************************
// Cells min reduction
∀ cells δt_courant <?= δt_cell_courant @ 12.11;
∀ cells δt_hydro   <?= δt_cell_hydro   @ 12.22;


void dumpTimeConstraintsForElems(void) @ 12.3{
  //debug()<<"\tdomain.dtcourant()="<<δt_courant;
  //debug()<<"\tdomain.dthydro()  ="<<δt_hydro;
  //printf("\n\tdomain.dtcourant()=%.21e",δt_courant);
  //printf("\n\tdomain.dthydro()  =%.21e",δt_hydro);
}



// ****************************************************************************
// * atexit: dumpFinalEnergy
// ****************************************************************************
/*void dumpFinalEnergy(void){
  ℝ MaxAbsDiff = 0.0;
  ℝ TotalAbsDiff = 0.0;
  ℝ MaxRelDiff = 0.0;
  int edgeElems=X_EDGE_ELEMS;
  debug()<<"   Problem size = "<<edgeElems;
  debug()<<"   Iteration count = "<<iteration-1;
  debug()<<"   Final Origin Energy = "<<e[0];
  debug()<<"";
  for(int j=0; j < edgeElems; ++j){
    for(int k=j+1; k < edgeElems; ++k){
      ℝ RelDiff,AbsDiff = fabs(e[j*edgeElems+k]-e[k*edgeElems+j]);
      TotalAbsDiff += AbsDiff;
      if (MaxAbsDiff<AbsDiff) MaxAbsDiff=AbsDiff;
      RelDiff = AbsDiff/e[k*edgeElems+j];
      if (MaxRelDiff<RelDiff)MaxRelDiff=RelDiff;
    }
  }
  debug()<<"\n   Testing Plane 0 of Energy Array:";
  debug()<<"        MaxAbsDiff   = "<<MaxAbsDiff;
  debug()<<"        TotalAbsDiff = "<<TotalAbsDiff;
  debug()<<"        MaxRelDiff   = "<<MaxRelDiff;
  }*/
